import os
from pathlib import Path
from unittest import mock

import boto3
import pytest
from assertpy import assert_that
from assertpy.assertpy import AssertionBuilder
from moto import mock_s3
from typing import Callable
from datetime import datetime
import time
from s3_url import S3Url

TEST_BUCKET = 'test-bucket'
DEFAULT_REGION = 'us-east-1'


@pytest.fixture(autouse=True, scope='function')
def aws_defaults():
    """Mocked AWS Credentials & Region for moto."""
    with mock.patch.dict(os.environ, {
        'AWS_ACCESS_KEY_ID': 'testing',
        'AWS_SECRET_ACCESS_KEY': 'testing',
        'AWS_SECURITY_TOKEN': 'testing',
        'AWS_SESSION_TOKEN': 'testing',
        'AWS_DEFAULT_REGION': DEFAULT_REGION
    }):
        yield


@pytest.fixture(autouse=True)
def s3_moto():
    with mock_s3():
        S3Url._s3_res = boto3.resource('s3')
        yield


@pytest.fixture
def s3_test_bucket(s3_moto):
    # just for demo - create 'inbound' bucket
    s3 = boto3.resource('s3')
    bucket = s3.create_bucket(Bucket=TEST_BUCKET)
    yield bucket


def s3_file_fixture(name, key, filename):
    @pytest.fixture
    def _s3_file_fixture(s3_test_bucket):
        current_path = Path(__file__).resolve().parent
        img_path = current_path / "resources" / filename
        with img_path.open(mode='rb') as fobj:
            s3_test_bucket.upload_fileobj(Fileobj=fobj, Key=key)
        yield key

    globals()[name] = _s3_file_fixture


s3_file_fixture('s3_demo_file', 'prefix/file.json', 'test_file.json')
s3_file_fixture('s3_test_file', 'SomeFolder/test_file.json', 'test_file.json')
s3_file_fixture('s3_test_file_2', 'SomeFolder/test_file_2.json', 'test_file.json')
s3_file_fixture('s3_test_file_3', 'a/path/to/the/file/test_file_3.json', 'test_file.json')


def assert_with_timeout(predicate: Callable, timeout_sec=90, poll_interval=1):
    started_at = datetime.now()
    while True:
        try:
            return predicate()
        except AssertionError as exc:
            time.sleep(poll_interval)
            if (datetime.now() - started_at).total_seconds() > timeout_sec:
                raise exc


def has_s3_tags_equal_to(self, expected_tags: dict):
    target_tags = S3Url(self.val).read_tags()
    assert_that(target_tags).is_equal_to(expected_tags)

    return self


def has_s3_tags_contains_entry(self, *args, **kwargs):
    target_tags = get_s3_url_obj(self).read_tags()
    assert_that(target_tags).contains_entry(*args, **kwargs)

    return self


def s3_file_exists(self):
    exists = get_s3_url_obj(self).exists()
    assert_that(exists).is_true()

    return self


def s3_file_does_not_exist(self):
    exists = get_s3_url_obj(self).exists()
    assert_that(exists).is_false()

    return self


def get_s3_url_obj(self) -> S3Url:
    return self.val if isinstance(self.val, S3Url) else S3Url(self.val)


AssertionBuilder.has_s3_tags_equal_to = has_s3_tags_equal_to
AssertionBuilder.has_s3_tags_contains_entry = has_s3_tags_contains_entry
AssertionBuilder.s3_file_exists = s3_file_exists
AssertionBuilder.s3_file_does_not_exist = s3_file_does_not_exist



import json
from pathlib import Path

import boto3
import pytest
import requests
from assertpy import assert_that
from boto3 import s3
from botocore.exceptions import ClientError
from s3_url import S3Url
from tests.conftest import TEST_BUCKET, assert_with_timeout


def test_demo(s3_test_bucket, s3_demo_file):
    file_url = S3Url(f's3://test-bucket/prefix/file.json')

    # url component properties
    assert file_url.bucket == 'test-bucket'
    assert file_url.key == 'prefix/file.json'

    # factory methods
    file_url = S3Url.from_url('s3://test-bucket/prefix/file.json')
    file_url = S3Url.from_bucket_key(bucket='test-bucket', key='prefix/file.json')

    # access underlying boto3 s3 resource object
    boto3_obj = file_url.object

    # return url string
    url_str: str = file_url.url
    url_str: str = str(file_url)
    assert url_str == 's3://test-bucket/prefix/file.json'

    # check if file exists
    exists: bool = file_url.exists()
    assert exists

    # check if any files exist in prefix (url should end with /)
    prefix_exists = S3Url('s3://test-bucket/prefix/').prefix_exists()
    assert prefix_exists

    # read text/json
    file_content_bin: bytes = file_url.read()
    file_content_text: str = file_url.read_text()
    file_content_json: json = file_url.read_json()

    # delete file
    file_url.delete()
    assert not file_url.exists()

    # write text/json
    file_url.write_text("test data")
    file_url.write_json({"testEntry": "test data"})
    assert file_url.exists()

    # copy to another object
    file_url.copy_to('s3://test-bucket/prefix/file-copy.json')
    S3Url('s3://test-bucket/prefix/another-file-copy.json').copy_from(file_url)

    # delete all filed in prefix
    prefix_url = S3Url('s3://test-bucket/prefix/')
    prefix_url.delete_dir()

    assert not prefix_url.prefix_exists()
    assert not file_url.exists()


def test_s3_url_parts(s3_test_bucket, s3_test_file):
    url = S3Url(f's3://{s3_test_bucket.name}/{s3_test_file}')
    assert_that(url.url).is_equal_to(f's3://{TEST_BUCKET}/SomeFolder/test_file.json')
    assert_that(url.bucket).is_equal_to(TEST_BUCKET)
    assert_that(url.key).is_equal_to('SomeFolder/test_file.json')


def test_factory_methods(s3_test_bucket, s3_test_file):
    url = S3Url.from_url(f's3://{s3_test_bucket.name}/{s3_test_file}')
    assert_that(url.url).is_equal_to(f's3://{TEST_BUCKET}/SomeFolder/test_file.json')
    assert_that(url.bucket).is_equal_to(TEST_BUCKET)
    assert_that(url.key).is_equal_to('SomeFolder/test_file.json')

    url = S3Url.from_bucket_key(s3_test_bucket.name, s3_test_file)
    assert_that(url.url).is_equal_to(f's3://{TEST_BUCKET}/SomeFolder/test_file.json')
    assert_that(url.bucket).is_equal_to(TEST_BUCKET)
    assert_that(url.key).is_equal_to('SomeFolder/test_file.json')


def test_unsupported_url(s3_test_bucket):
    with pytest.raises(Exception) as err:
        S3Url(f'https://{s3_test_bucket.name}/some/path')
    assert_that(str(err)).contains('Unsupported URL')


def test_s3_url_exists(s3_test_bucket, s3_test_file):
    existing_url = S3Url(f's3://{s3_test_bucket.name}/{s3_test_file}')
    assert_that(existing_url.exists()).is_true()
    non_existing_url = S3Url(f's3://{s3_test_bucket.name}/non_existing_file.txt')
    assert_that(non_existing_url.exists()).is_false()


def test_generate_presigned_url_get(s3_test_bucket, s3_test_file):
    existing_url = S3Url(f's3://{s3_test_bucket.name}/{s3_test_file}')
    get_url = existing_url.generate_presigned_url_get()
    assert_that(requests.get(get_url).content.decode()).is_equal_to(existing_url.read_text())


def test_generate_presigned_url_put(s3_test_bucket, s3_test_file):
    non_existing_url = S3Url(f's3://{s3_test_bucket.name}/non_existing_file.txt')
    get_url = non_existing_url.generate_presigned_url_put()
    requests.put(get_url, data="test text file".encode()).raise_for_status()
    assert_that(non_existing_url.read_text()).is_equal_to("test text file")


def test_s3_url_exists_access_denied(s3_test_bucket, s3_test_file, monkeypatch):
    existing_url = S3Url(f's3://{s3_test_bucket.name}/{s3_test_file}')
    expected_exc = ClientError({'Error': {'Code': '403', 'Message': 'Forbidden'},
                                'ResponseMetadata': {'HTTPStatusCode': 403, 'RetryAttempts': 0}}, 'HeadObject')

    def raise_exc(*args, **kwargs):
        raise expected_exc

    monkeypatch.setattr(existing_url.object.meta.client, '_make_api_call', raise_exc)
    assert_that(existing_url.exists).raises(ClientError).when_called_with()


def test_s3_url_exists_for_a_path(s3_test_bucket, s3_test_file_3):
    existing_url = S3Url(f's3://{s3_test_bucket.name}/a/path/to/')
    assert_that(existing_url.prefix_exists()).is_true()
    existing_url = S3Url(f's3://{s3_test_bucket.name}/a/path/to/the/file')
    assert_that(existing_url.prefix_exists()).is_true()

    non_existing_url = S3Url(f's3://{s3_test_bucket.name}/unknown/path/to/the/file')
    assert_that(non_existing_url.prefix_exists()).is_false()


def test_list_prefix_objects(s3_test_bucket):
    file_1 = S3Url(f's3://{s3_test_bucket.name}/some_prefix/some_file_1.txt')
    file_2 = S3Url(f's3://{s3_test_bucket.name}/some_prefix/some_file_2.txt')
    file_3 = S3Url(f's3://{s3_test_bucket.name}/some_prefix/some_file_3.txt')

    file_1.write_text("1")
    file_2.write_text("2")
    file_3.write_text("3")

    existing_prefix_url = S3Url(f's3://{s3_test_bucket.name}/some_prefix/')
    assert_that(list(existing_prefix_url.list_prefix_objects())).is_length(3).contains_only(file_1, file_2, file_3)


def test_list_prefix_objects_non_existing_prefix(s3_test_bucket):
    existing_prefix_url = S3Url(f's3://{s3_test_bucket.name}/non-existing/')
    assert_that(list(existing_prefix_url.list_prefix_objects())).is_empty()


def test_list_common_prefixes(s3_test_bucket):
    S3Url(f's3://{s3_test_bucket.name}/some_prefix/sub1/some_file_1.txt').write_text("test")
    S3Url(f's3://{s3_test_bucket.name}/some_prefix/sub1/some_file_2.txt').write_text("test")
    S3Url(f's3://{s3_test_bucket.name}/some_prefix/sub2/some_file_3.txt').write_text("test")
    S3Url(f's3://{s3_test_bucket.name}/some_prefix/sub3/sub4/some_file_3.txt').write_text("test")
    assert_that(list(S3Url(f's3://{s3_test_bucket.name}/some_prefix/sub1/').list_common_prefixes())).is_empty()
    assert_that(list(S3Url(f's3://{s3_test_bucket.name}/some_prefix/sub2/').list_common_prefixes())).is_empty()
    assert_that(list(S3Url(f's3://{s3_test_bucket.name}/some_prefix/sub3/').list_common_prefixes())).is_length(1) \
        .contains(S3Url('s3://test-bucket/some_prefix/sub3/sub4/'))
    assert_that(list(S3Url(f's3://{s3_test_bucket.name}/some_prefix/').list_common_prefixes())) \
        .is_length(3) \
        .contains(
        S3Url('s3://test-bucket/some_prefix/sub1/'),
        S3Url('s3://test-bucket/some_prefix/sub2/'),
        S3Url('s3://test-bucket/some_prefix/sub3/'),
    )


def test_list_common_prefixes_empty(s3_test_bucket):
    res = list(S3Url(f's3://{s3_test_bucket.name}/non-existing/').list_common_prefixes())
    assert_that(res).is_empty()


def test_s3_url_exists_for_a_path_access_denied(s3_test_bucket, s3_test_file_3, monkeypatch):
    existing_url = S3Url(f's3://{s3_test_bucket.name}/a/path/to/')
    expected_exc = ClientError({'Error': {'Code': '403', 'Message': 'Forbidden'},
                                'ResponseMetadata': {'HTTPStatusCode': 403, 'RetryAttempts': 0}}, 'ListObjects')

    def raise_exc(*args, **kwargs):
        raise expected_exc

    monkeypatch.setattr(existing_url.object.meta.client, '_make_api_call', raise_exc)
    assert_that(existing_url.prefix_exists).raises(ClientError).when_called_with()


def test_s3_url_delete(s3_test_bucket, s3_test_file):
    existing_url = S3Url(f's3://{s3_test_bucket.name}/{s3_test_file}')
    assert_that(existing_url.exists()).is_true()
    existing_url.delete()
    assert_that(existing_url.exists()).is_false()
    # should not fail if not exists
    existing_url.delete()


def test_s3_url_delete_dir(s3_test_bucket, s3_test_file):
    existing_url = S3Url(f's3://{s3_test_bucket.name}/{s3_test_file}')
    assert_that(existing_url.exists()).is_true()
    prefix = S3Url(f's3://{s3_test_bucket.name}/SomeFolder/')
    prefix.delete_dir()
    assert_that(existing_url.exists()).is_false()


def test_s3_url_read_write(s3_test_bucket, s3_test_file):
    url = S3Url(f's3://{s3_test_bucket.name}/{s3_test_file}')
    assert_that(url.read()).is_equal_to(b'{\n  "testEntry1": "value1"\n}')
    assert_that(json.loads(url.read_text())).is_equal_to({
        "testEntry1": "value1"
    })
    url.write_text(json.dumps({
        "newField": "newValue"
    }))
    assert_that(url.object.server_side_encryption).is_none()

    assert_that(json.loads(url.read_text())).is_equal_to({
        "newField": "newValue"
    })
    url.write_text(json.dumps({
        "newField": "newValue"
    }), encryption='AES256')
    assert_that(url.object.server_side_encryption).is_equal_to('AES256')


def test_s3_url_read_write_json(s3_test_bucket, s3_test_file):
    url = S3Url(f's3://{s3_test_bucket.name}/{s3_test_file}')
    assert_that(url.read_json()).is_equal_to({
        "testEntry1": "value1"
    })
    url.write_json({
        "newField": "newValue"
    })
    assert_that(url.object.server_side_encryption).is_none()

    assert_that(url.read_json()).is_equal_to({
        "newField": "newValue"
    })
    url.write_json({
        "newField": "newValue"
    }, encryption='AES256')
    assert_that(url.object.server_side_encryption).is_equal_to('AES256')


def test_s3_url_read_non_existing(s3_test_bucket):
    non_existing_url = S3Url(f's3://{s3_test_bucket.name}/non_existing_file.txt')
    with pytest.raises(ClientError) as err:
        non_existing_url.read_text()
    assert_that(str(err.value)).contains('NoSuchKey')
    assert_that(str(err.value)).contains('The specified key does not exist')


def test_s3_url_upload_file(s3_test_bucket):
    url = S3Url(f's3://{s3_test_bucket.name}/some_new_file.json')
    current_path = Path(__file__).resolve().parent
    img_path = current_path / "resources" / "test_file.json"
    with img_path.open(mode='rb') as fobj:
        url.upload_file(fobj)
    assert_that(json.loads(url.read_text())).is_equal_to({
        "testEntry1": "value1"
    })
    assert_that(url.object.server_side_encryption).is_none()


def test_s3_url_upload_file_encryption(s3_test_bucket):
    url = S3Url(f's3://{s3_test_bucket.name}/some_new_file.json')
    current_path = Path(__file__).resolve().parent
    img_path = current_path / "resources" / "test_file.json"
    with img_path.open(mode='rb') as fobj:
        url.upload_file(fobj, encryption='AES256')
    assert_that(json.loads(url.read_text())).is_equal_to({
        "testEntry1": "value1"
    })
    assert_that(url.object.server_side_encryption).is_equal_to('AES256')


def test_copy_to_url(s3_test_bucket, s3_test_file):
    # Given
    source_url = S3Url(f's3://{s3_test_bucket.name}/{s3_test_file}')
    target_url = S3Url(f's3://{s3_test_bucket.name}/copied/{s3_test_file}')
    assert_that(target_url).s3_file_does_not_exist()

    # When
    source_url.copy_to(target_url)

    # Then
    assert_that(target_url).s3_file_exists()


def test_copy_to_str(s3_test_bucket, s3_test_file):
    # Given
    source_url = S3Url(f's3://{s3_test_bucket.name}/{s3_test_file}')
    target_url = S3Url(f's3://{s3_test_bucket.name}/copied/{s3_test_file}')
    assert_that(target_url).s3_file_does_not_exist()

    # When
    source_url.copy_to(target_url.url)

    # Then
    assert_that(target_url).s3_file_exists()


def test_copy_from_url(s3_test_bucket, s3_test_file):
    # Given
    source_url = S3Url(f's3://{s3_test_bucket.name}/{s3_test_file}')
    target_url = S3Url(f's3://{s3_test_bucket.name}/copied/{s3_test_file}')
    assert_that(target_url).s3_file_does_not_exist()

    # When
    target_url.copy_from(source_url)

    # Then
    assert_that(target_url).s3_file_exists()


def test_copy_from_str(s3_test_bucket, s3_test_file):
    # Given
    source_url = S3Url(f's3://{s3_test_bucket.name}/{s3_test_file}')
    target_url = S3Url(f's3://{s3_test_bucket.name}/copied/{s3_test_file}')
    assert_that(target_url).s3_file_does_not_exist()

    # When
    target_url.copy_from(source_url.url)

    # Then
    assert_that(target_url).s3_file_exists()


def test_should_read_tags_to_s3_file(s3_test_bucket, s3_test_file):
    # Given
    url = S3Url(f's3://{s3_test_bucket.name}/{s3_test_file}')
    s3_test_bucket.meta.client.put_object_tagging(
        Bucket=s3_test_bucket.name,
        Key=s3_test_file,
        Tagging={
            'TagSet': [
                {'Key': 'tag1', 'Value': 'value1'},
                {'Key': 'tag2', 'Value': 'value2'},
            ]
        }
    )

    # Then
    assert_that(url).has_s3_tags_equal_to({'tag1': 'value1', 'tag2': 'value2'})


def test_should_add_and_read_no_tags_to_s3_file(s3_test_bucket, s3_test_file):
    # Given
    input_tags = {}
    url = S3Url(f's3://{s3_test_bucket.name}/{s3_test_file}')

    # When
    url.write_tags(input_tags)

    # Then
    assert_that(url).has_s3_tags_equal_to({})


def test_should_add_and_read_none_tags_to_s3_file(s3_test_bucket, s3_test_file):
    # Given
    input_tags = None

    url = S3Url(f's3://{s3_test_bucket.name}/{s3_test_file}')

    # When
    url.write_tags(input_tags)

    # Then
    assert_that(url).has_s3_tags_equal_to({})


def test_should_copy_s3_file_with_tags(s3_test_bucket, s3_test_file):
    # Given
    source_s3_url = S3Url(f's3://{s3_test_bucket.name}/{s3_test_file}')
    target_s3_url = S3Url(f's3://{s3_test_bucket.name}/copied/{s3_test_file}')
    s3_test_bucket.meta.client.put_object_tagging(
        Bucket=s3_test_bucket.name,
        Key=s3_test_file,
        Tagging={
            'TagSet': [
                {'Key': 'tag1', 'Value': 'value1'},
                {'Key': 'tag2', 'Value': 'value2'},
            ]
        }
    )

    # When
    source_s3_url.copy_to(target_s3_url)

    # Then
    assert_that(target_s3_url).s3_file_exists()
    assert_that(target_s3_url).has_s3_tags_equal_to({'tag1': 'value1', 'tag2': 'value2'})


def test_should_copy_tags_to_url(s3_test_bucket, s3_test_file, s3_test_file_2):
    # Given
    source_url = S3Url(f's3://{s3_test_bucket.name}/{s3_test_file}')
    target_url = S3Url(f's3://{s3_test_bucket.name}/{s3_test_file_2}')

    assert_that(source_url).has_s3_tags_equal_to({})
    assert_that(target_url).has_s3_tags_equal_to({})

    # When
    source_url.write_tags({'tag1': 'value1', 'tag2': 'value2'})
    source_url.copy_tags_to(target_url)

    # Then
    assert_that(target_url).s3_file_exists()
    assert_that(target_url).has_s3_tags_equal_to({'tag1': 'value1', 'tag2': 'value2'})


def test_should_copy_tags_to_str(s3_test_bucket, s3_test_file, s3_test_file_2):
    # Given
    source_url = S3Url(f's3://{s3_test_bucket.name}/{s3_test_file}')
    target_url = S3Url(f's3://{s3_test_bucket.name}/{s3_test_file_2}')

    assert_that(source_url).has_s3_tags_equal_to({})
    assert_that(target_url).has_s3_tags_equal_to({})

    # When
    source_url.write_tags({'tag1': 'value1', 'tag2': 'value2'})
    source_url.copy_tags_to(target_url.url)

    # Then
    assert_that(target_url).has_s3_tags_equal_to({'tag1': 'value1', 'tag2': 'value2'})


def test_should_copy_s3_file_with_no_tags(s3_test_bucket, s3_test_file, s3_test_file_2):
    # Given
    source_url = S3Url(f's3://{s3_test_bucket.name}/{s3_test_file}')
    target_url = S3Url(f's3://{s3_test_bucket.name}/{s3_test_file_2}')

    assert_that(source_url).has_s3_tags_equal_to({})
    assert_that(target_url).has_s3_tags_equal_to({})

    # When
    source_url.copy_tags_to(target_url)

    # Then
    assert_that(target_url.exists()).is_true()
    assert_that(target_url).has_s3_tags_equal_to({})


def test_transition_to_glacier(s3_test_bucket, s3_test_file):
    # Given
    source_url = S3Url(f's3://{s3_test_bucket.name}/{s3_test_file}')

    # When
    source_url.transition_to_storage_tier("GLACIER")

    # Then
    assert_that(source_url.object.storage_class).is_equal_to("GLACIER")


def test_shouldnt_transition_from_glacier(s3_test_bucket, s3_test_file):
    # Given
    source_url = S3Url(f's3://{s3_test_bucket.name}/{s3_test_file}')
    source_url.transition_to_storage_tier("GLACIER")

    # When
    with pytest.raises(Exception):
        source_url.transition_to_storage_tier("STANDARD")


def test_transition_from_int_tiering(s3_test_bucket, s3_test_file):
    # Given
    source_url = S3Url(f's3://{s3_test_bucket.name}/{s3_test_file}')
    source_url.transition_to_storage_tier("INTELLIGENT_TIERING")

    # When
    source_url.transition_to_storage_tier("STANDARD")


def test_transition_from_onezone(s3_test_bucket, s3_test_file):
    # Given
    source_url = S3Url(f's3://{s3_test_bucket.name}/{s3_test_file}')
    source_url.transition_to_storage_tier("ONEZONE_IA")

    # When
    source_url.transition_to_storage_tier("STANDARD")


def test_transition_to_storage_tier(s3_test_bucket, s3_test_file):
    # Given
    source_url = S3Url(f's3://{s3_test_bucket.name}/{s3_test_file}')

    # When
    source_url.transition_to_storage_tier("STANDARD_IA")

    # Then
    assert_that(source_url.object.storage_class).is_equal_to("STANDARD_IA")


def test_restore_from_glacier(s3_test_bucket, s3_test_file):
    # Given
    source_url = S3Url(f's3://{s3_test_bucket.name}/{s3_test_file}')
    source_url.transition_to_storage_tier("GLACIER")

    # When
    source_url.restore_to_storage_tier(1)

    def assert_object_restored():
        obj = source_url.object
        assert_that(obj.restore).is_not_none().starts_with('ongoing-request="false"')

    assert_with_timeout(assert_object_restored, 100, 1)
    assert_that(source_url.object.storage_class).is_equal_to("GLACIER")
    assert_that(source_url.object.restore).is_not_none().starts_with('ongoing-request="false"')

    # When
    source_url.restore_to_storage_tier(1, "Expedited")

    assert_with_timeout(assert_object_restored, 100, 1)
    assert_that(source_url.object.storage_class).is_equal_to("GLACIER")



from io import IOBase
import json
import threading
from pathlib import Path
from typing import Union, Iterable, IO
from urllib.parse import urlparse

import boto3
from botocore.exceptions import ClientError


class S3Url():
    _local = threading.local()

    def __init__(self, url: Union[str, 'S3Url']):
        '''
        Creates object instance
        :param url: "s3://"-shaped url or another S3Url object
        '''

        if not hasattr(self._local, 's3_res'):
            self._local.s3_res = boto3.resource('s3')
        if isinstance(url, S3Url):
            url = url.url
        else:
            if not url.startswith('s3://'):
                raise Exception(f'Unsupported URL: {url}. It must start with s3://')
        self._parsed = urlparse(url, allow_fragments=False)
        self._object = self._local.s3_res.Object(self.bucket, self.key)

    @classmethod
    def from_url(cls, url: Union[str, 'S3Url']) -> 'S3Url':
        return S3Url(url)

    @classmethod
    def from_bucket_key(cls, bucket: str, key: str) -> 'S3Url':
        return S3Url(f's3://{bucket}/{key}')

    def __repr__(self) -> str:
        return self.url

    def __eq__(self, o: object) -> bool:
        return isinstance(o, S3Url) and o.url == self.url

    def __hash__(self) -> int:
        return self.url.__hash__()

    @property
    def bucket(self) -> str:
        return self._parsed.netloc

    @property
    def key(self) -> str:
        return self._parsed.path.lstrip('/')

    @property
    def object(self):
        return self._object

    @property
    def url(self) -> str:
        return self._parsed.geturl()

    def exists(self) -> bool:
        try:
            self.object.load()
            return True
        except ClientError as cerr:
            if '404' in str(cerr):
                return False
            else:
                raise cerr

    def prefix_exists(self) -> bool:
        try:
            next(iter(self._object.Bucket().objects.filter(Prefix=self.key)))
            return True
        except StopIteration:
            return False
        except ClientError as cerr:
            if '404' in str(cerr):
                return False
            else:
                raise cerr

    def read_text(self, encoding="utf-8-sig") -> str:
        return self.read().decode(encoding)

    def read(self) -> bytes:
        return self.object.get()['Body'].read()

    def read_json(self, encoding="utf-8-sig") -> json:
        return json.loads(self.read_text(encoding))

    def delete(self) -> None:
        self.object.delete()

    def write(self, body: Union[str, bytes], encryption=None) -> None:
        if encryption:
            self.object.put(Body=body, ServerSideEncryption=encryption)
        else:
            self.object.put(Body=body)

    def write_text(self, body: str, encryption=None) -> None:
        self.write(body, encryption)

    def write_json(self, body: json, encryption=None) -> None:
        self.write(json.dumps(body), encryption)

    def upload_file(self, fileobj: IO, encryption=None):
        if encryption:
            self.object.upload_fileobj(fileobj, ExtraArgs={
                'ServerSideEncryption': encryption
            })
        else:
            self.object.upload_fileobj(fileobj)

    def delete_dir(self):
        for obj in self._local.s3_res.Bucket(self.bucket).objects.filter(Prefix=self.key):
            obj.delete()

    def write_tags(self, tags: dict) -> None:
        if tags:
            tag_set = [{'Key': k, 'Value': v} for k, v in tags.items()]
            self._local.s3_res.meta.client.put_object_tagging(
                Bucket=self.bucket,
                Key=self.key,
                Tagging={
                    'TagSet': tag_set
                }
            )

    def read_tags(self) -> dict:
        tags_dict = {}
        tags = self._local.s3_res.meta.client.get_object_tagging(
            Bucket=self.bucket,
            Key=self.key,
        )
        if tags:
            tags_dict = {x['Key']: x['Value'] for x in tags['TagSet']}

        return tags_dict

    def transition_to_storage_tier(self, storage_tier: str):
        return self._local.s3_res.meta.client.copy_object(
            CopySource={
                'Bucket': self.bucket,
                'Key': self.key
            },
            Bucket=self.bucket,
            Key=self.key,
            StorageClass=storage_tier,
            MetadataDirective='COPY')

    def restore_to_storage_tier(self, days: int, retrieval_tier: str = "Standard"):
        return self._local.s3_res.meta.client.restore_object(
            Bucket=self.bucket,
            Key=self.key,
            RestoreRequest={'Days': days, 'GlacierJobParameters': {'Tier': retrieval_tier}})

    def copy_to(self, target_url: Union[str, 'S3Url']) -> None:
        if isinstance(target_url, S3Url):
            target_obj = target_url
        else:
            target_obj = S3Url(target_url)
        self._local.s3_res.meta.client.copy({
            'Bucket': self.bucket,
            'Key': self.key
        }, target_obj.bucket, target_obj.key)

    def copy_from(self, source_url: Union[str, 'S3Url']) -> None:
        if isinstance(source_url, S3Url):
            source_obj = source_url
        else:
            source_obj = S3Url(source_url)
        self._local.s3_res.meta.client.copy({
            'Bucket': source_obj.bucket,
            'Key': source_obj.key
        }, self.bucket, self.key)

    def copy_tags_to(self, target_url: Union[str, 'S3Url']) -> None:
        source_tags = self.read_tags()
        if isinstance(target_url, S3Url):
            target_obj = target_url
        else:
            target_obj = S3Url(target_url)
        target_obj.write_tags(source_tags)

    def copy_tags_from(self, source_url: Union[str, 'S3Url']) -> None:
        if isinstance(source_url, S3Url):
            source_obj = source_url
        else:
            source_obj = S3Url(source_url)
        source_tags = source_obj.read_tags()
        self.write_tags(source_tags)

    def list_prefix_objects(self) -> Iterable['S3Url']:
        for s3_obj in self._object.Bucket().objects.filter(Prefix=self.key):
            yield S3Url(f's3://{s3_obj.bucket_name}/{s3_obj.key}')

    def list_common_prefixes(self) -> Iterable['S3Url']:
        # for prefix in self._local.s3_res.meta.client.list_objects(Bucket=self.bucket, Prefix=self.key, Delimiter='/')[
        for prefix in self._local.s3_res.meta.client\
                .get_paginator('list_objects')\
                .paginate(Bucket=self.bucket, Prefix=self.key, Delimiter='/').search('CommonPrefixes'):
            if prefix:
                yield S3Url(f's3://{self.bucket}/{prefix["Prefix"]}')

    def generate_presigned_url_get(self, timeout=3600) -> str:
        return self._local.s3_res.meta.client.generate_presigned_url(
            ClientMethod='get_object',
            Params={'Bucket': self.bucket, 'Key': self.key},
            ExpiresIn=timeout
        )

    def generate_presigned_url_put(self, timeout=3600) -> str:
        return self._local.s3_res.meta.client.generate_presigned_url(
            ClientMethod='put_object',
            Params={'Bucket': self.bucket, 'Key': self.key},
            ExpiresIn=timeout
        )
